#!/usr/bin/env python3
import os
from glob import glob
import arrow
from pathlib import Path
import calendar
import humanize
import argparse
import pathlib
import logging
import datetime as dt
import click
from . import cli, pass_config, Commands
from .tools import abort

DTF = "%Y-%m-%d %H:%M:%S"


log = logging.getLogger()


def _get_files_if_dir(path):
    for path in glob(str(path)):
        path = Path(path)
        if not path.name:
            continue
        if path.is_dir():
            for file in path.glob("*"):
                if file.is_file():
                    yield next(_get_files_if_dir(file))
        elif path.is_file():
            mtime = arrow.get(path.stat().st_mtime)
            is_dir = path.is_dir()
            is_file = path.is_file()
            log.debug(f"{path}: mtime {mtime} dir? {is_dir} file? {is_file}")
            yield (path, mtime)


def genPathInfos(arg_paths):

    for arg_path in arg_paths:
        po = pathlib.Path(arg_path)
        log.debug(po)
        yield from _get_files_if_dir(arg_path)


def rm(path_list, dry_run):
    now = arrow.utcnow()
    log.info("Starting deletion:")
    for path in path_list:
        if path.is_file():
            if dry_run:
                modified = arrow.get(path.stat().st_mtime)

                click.secho(
                    "dry run -- would delete:"
                    f"{modified.strftime(DTF)}"
                    f" {path} "
                    f" {humanize.naturaldelta(now - modified)}"
                )
            else:
                path.unlink()
                log.info(f"Deleted: {path}")


def get_bins():
    def _return(x, y):
        return _hour_0(x), _hour_235959(y)

    def _hour_0(d):
        return d.replace(hour=0, minute=0, second=0)

    def _hour_235959(d):
        return d.replace(hour=23, minute=59, second=59)

    winning_weekday = 6  # sunday
    start = arrow.get()
    while start.weekday() != winning_weekday:
        start = start.shift(days=-1)

    # last x weeks
    def get_weeks():
        X = start
        for i in range(4):
            yield _return(X.shift(days=-6), X)
            X = X.shift(weeks=-1)

    def get_months():
        X = start
        for i in range(6):
            X = X.shift(months=-1)
            last_of_month = calendar.monthrange(X.year, X.month)[1]
            yield _return(
                arrow.get(X.year, X.month, 1),
                arrow.get(X.year, X.month, last_of_month),
            )

    def get_quarters():
        X = start
        for i in range(12):
            X = X.shift(months=-1)
            if X.month not in (3, 6, 9, 12):
                continue
            last_of_month = calendar.monthrange(X.year, X.month)[1]
            yield _return(
                arrow.get(X.year, X.month, 1),
                arrow.get(X.year, X.month, last_of_month),
            )

    def get_years():
        for i in range(20):
            year_end = arrow.get().replace(year=arrow.get().year - i, month=12, day=31)
            yield _return(
                year_end.replace(day=1, month=1),
                year_end,
            )

    yield _return(start, start)
    yield from get_weeks()
    yield from get_months()
    yield from get_quarters()
    yield from get_years()


def get_to_delete_files(path_list, days_notouch):
    now = dt.datetime.utcnow()
    log.debug(f"Now: {now}")
    bins = {}
    for bin in set(get_bins()):
        bins.setdefault(bin, [])

    keep_safe = set()
    to_delete = []
    for path, mt in genPathInfos(path_list):
        if (arrow.utcnow() - mt).days <= days_notouch:
            print("Ignoring", path)
            keep_safe.add(path)
            continue

        for key in bins:
            if mt >= key[0] and mt <= key[1]:
                bins[key].append(path)
        else:
            to_delete.append(path)

    # sort arrays by date reverse; at position[0] is the file
    # that will survive
    for k in bins.keys():
        bins[k] = sorted(bins[k], key=lambda x: x.stat().st_mtime, reverse=True)

    # store pole positions in safe array
    [keep_safe.add(x[0]) for x in bins.values() if x]

    # collect all victims
    for files in bins.values():
        to_delete += files[1:]

    click.secho("==========================")
    click.secho("Kept:\n")
    size = print_files(sorted(keep_safe))
    click.secho(f"Keeping {humanize.naturalsize(size)}")
    click.secho("==========================")

    return list(set(to_delete) - keep_safe)


def print_files(files):
    size = 0
    for path in list(set(files)):
        size += path.stat().st_size
        date = arrow.get(path.stat().st_mtime).strftime("%Y-%m-%d %H:%M:%S")
        click.secho(f"{path}: {date}")
    return size


@cli.command(
    help=(
        "Deletes file matching the given glob in PATH and keeps "
        "youngest files of last weeks, months, quarters and years. "
        "By providing --doNt-touch files can be provided, that are "
        "never touched. "
    )
)
@click.argument("path", required=True, nargs=-1)
@click.option(
    "-n",
    "--dry-run",
    is_flag=True,
    help="Do not touch the last X days from today. Defaults to 1=yesterday",
)
@click.option(
    "-N",
    "--dont-touch",
    default=1,
    help="Do not touch the last X days from today. Defaults to 1=yesterday",
)
@pass_config
def daddy_cleanup(config, path, dry_run, dont_touch):
    if config.verbose:
        log.setLevel(logging.DEBUG)
    path = [Path(os.getcwd()) / x for x in path]
    deletion_candidates = list(sorted(set(get_to_delete_files(path, dont_touch))))
    if deletion_candidates:
        size = print_files(deletion_candidates)
        print("Going to delete ", humanize.naturalsize(size))
        rm(deletion_candidates, dry_run=dry_run)


@cli.command()
@click.argument("path", required=True, nargs=-1)
@click.option(
    "-n",
    "--dry-run",
    is_flag=True,
    help="Do not touch the last X days from today. Defaults to 1=yesterday",
)
@click.option(
    "-N",
    "--dont-touch",
    default=1,
    help="Do not touch the last X days from today. Defaults to 1=yesterday",
)
@pass_config
def keep_last_file_of_day(config, path, dry_run, dont_touch):
    if config.verbose:
        log.setLevel(logging.DEBUG)
    files = []
    for path in path:
        for file, mtime in _get_files_if_dir(path):
            if (arrow.utcnow() - mtime).days <= dont_touch:
                continue
            files.append((file, mtime))
    if not files:
        abort("No files matching")
    size = 0
    files = sorted(files, key=lambda x: x[1], reverse=False)
    for file, mtime in files[:-1]:
        size += file.stat().st_size
        if dry_run:
            click.secho(f"Would delete: {file}", fg="yellow")
        else:
            file.unlink()
            click.secho(f"Deleted: {file}", fg="red")
    click.secho(f"Keeping: {files[-1][0]}", fg="green")
    click.secho(
        f"{'Theoretically' if dry_run else ''} Deleted size: {humanize.naturalsize(size)}"
    )



import sys
import subprocess
from datetime import datetime
from pathlib import Path
from .click_config import Config
import imp
import inspect
import os
import shellingham

# from .myconfigparser import MyConfigParser  # NOQA load this module here, otherwise following lines and sublines get error
from .init_functions import load_dynamic_modules
from .init_functions import _get_customs_root

from .click_global_commands import GlobalCommands

try:
    import click
    from .lib_clickhelpers import AliasedGroup
except ImportError:
    click = None
from .tools import _file2env

from . import module_tools  # NOQA
from . import odoo_config  # NOQA

SCRIPT_DIRECTORY = Path(inspect.getfile(inspect.currentframe())).absolute().parent

Commands = GlobalCommands()

os.environ["HOST_HOME"] = os.getenv("HOME", "")
os.environ["ODOO_HOME"] = str(SCRIPT_DIRECTORY)


class NoProjectNameException(Exception):
    pass


def _get_default_project_name(restrict):
    def _get_project_name_from_file(path):
        if not path.exists():
            return
        pj = [x for x in path.read_text().split("\n") if "PROJECT_NAME" in x]
        if pj:
            return pj[0].split("=")[-1].strip()

    if restrict:
        paths = restrict
    else:
        paths = [Path(os.path.expanduser("~/.odoo/settings"))]

    for path in paths:
        pj = _get_project_name_from_file(path)
        if pj:
            return pj

    customs_root = _get_customs_root(Path(os.getcwd()))
    if customs_root:
        root = Path(customs_root)
        if (root / "MANIFEST").exists():
            return root.name
    raise NoProjectNameException("No default project name could be determined.")


pass_config = click.make_pass_decorator(Config, ensure=True)


@click.group(cls=AliasedGroup)
@click.option("-f", "--force", is_flag=True)
@click.option("-v", "--verbose", is_flag=True)
@click.option("--version", is_flag=True)
@click.option(
    "-xs",
    "--restrict-setting",
    multiple=True,
    help="Several parameters; limit to special configuration files settings and docker-compose files. All other configuration files will be ignored.",
)
@click.option(
    "-xd",
    "--restrict-docker-compose",
    multiple=True,
    help="Several parameters; limit to special configuration files settings and docker-compose files. All other configuration files will be ignored.",
)
@click.option("-p", "--project-name", help="Set Project-Name")
@click.option("--chdir", help="Set Working Directory")
@pass_config
def cli(
    config,
    force,
    verbose,
    project_name,
    restrict_setting,
    restrict_docker_compose,
    chdir,
    version,
):
    config.force = force
    config.verbose = verbose
    config.restrict = {}
    if verbose:
        os.environ["WODOO_VERBOSE"] = "1"
    if chdir:
        os.chdir(chdir)
        config.WORKING_DIR = chdir
    # if not config.WORKING_DIR:
    #     # usually all need a working except cicd
    #     if not any(x in sys.argv for x in ['init', 'install-completion']):
    #         click.secho("Please enter into an odoo directory, which contains a MANIFEST file.", fg='red')
    #         sys.exit(1)

    def _collect_files(files):
        for test in files:
            test = Path(test)
            if not test.exists():
                click.secho(f"Not found: {test}", fg="red")
                sys.exit(-1)
            yield test.absolute()

    config.restrict["settings"] = list(_collect_files(restrict_setting))
    config.restrict["docker-compose"] = list(_collect_files(restrict_docker_compose))

    if project_name:
        config.project_name = project_name
    else:
        try:
            config.project_name = _get_default_project_name(config.restrict["settings"])
        except NoProjectNameException:
            config.project_name = ""
    os.environ["project_name"] = config.project_name
    os.environ["docker_compose"] = str(config.files.get("docker_compose")) or ""
    os.environ["CUSTOMS_DIR"] = str(config.WORKING_DIR)

    load_dynamic_modules(config.dirs["images"])

    if config.verbose:
        print(config.files["docker_compose"])


from . import lib_clickhelpers  # NOQA
from . import lib_composer  # NOQA
from . import lib_backup  # NOQA
from . import lib_control  # NOQA
from . import lib_db  # NOQA
from . import lib_db_snapshots  # NOQA
from . import lib_lang  # NOQA
from . import lib_module  # NOQA
from . import lib_setup  # NOQA
from . import lib_src  # NOQA
from . import lib_docker_registry  # NOQA
from . import lib_venv  # NOQA
from . import lib_turnintodev  # NOQA
from . import daddy_cleanup  # NOQA

# import container specific commands
from .tools import abort  # NOQA
from .tools import __dcrun  # NOQA
from .tools import __dc  # NOQA


@cli.command()
@click.option(
    "-x",
    "--execute",
    is_flag=True,
    help=("Execute the script to insert completion into users rc-file."),
)
def completion(execute):
    shell = os.environ["SHELL"].split("/")[-1]
    rc_file = Path(os.path.expanduser(f"~/.{shell}rc"))
    line = f'eval "$(_ODOO_COMPLETE={shell}_source odoo)"'
    if execute:
        content = rc_file.read_text().splitlines()
        if not list(
            filter(
                lambda x: line in x and not x.strip().startswith("#"),
                content,
            )
        ):
            content += [f"\n{line}\n"]
            click.secho(
                f"Inserted successfully\n{line}" "\n\nPlease restart you shell."
            )
            rc_file.write_text("\n".join(content))
        else:
            click.secho("Nothing done - already existed.")
    else:
        click.secho(
            "\n\n" f"Insert into {rc_file}\n\n" f"echo '{line}' >> {rc_file}" "\n\n"
        )
    sys.exit(0)


@cli.command()
@pass_config
def version(config):
    from .tools import _get_version

    version = _get_version()

    images_sha = subprocess.check_output(
        ["git", "log", "-n1", "--format=%H"], encoding="utf8", cwd=config.dirs["images"]
    ).strip()
    images_branch = subprocess.check_output(
        ["git", "rev-parse", "--abbrev-ref", "HEAD"],
        encoding="utf8",
        cwd=config.dirs["images"],
    ).strip()
    click.secho(
        (
            f"Wodoo Version:    {version}\n"
            f"Images SHA:       {images_sha}\n"
            f"Images Branch:    {images_branch}\n"
        ),
        fg="yellow",
    )
